#if 0

Copyright (c) 2018, Christoph Schied
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Karlsruhe Institute of Technology nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#endif
#include "ReconstructionFilter.h"

extern int gradientDownsample;

class A_SVGF
    : public ReconstructionFilter
    , public CreateReconstructionFilter<A_SVGF>
    , public std::enable_shared_from_this<A_SVGF>
{
public:
    enum {
        ACCUM_COLOR,
        ACCUM_MOMENTS,
        ACCUM_LENGTH,
    };

    A_SVGF(RTContext::SharedPtr& pRTContext, RenderContext::SharedPtr& pRenderContext, const std::string& groupName, Gui::UniquePtr& pGui)
        : ReconstructionFilter(pRTContext, pRenderContext, groupName, pGui)
    {
        setGuiGroupName(groupName);
    }
    using SharedPtr = std::shared_ptr<A_SVGF>;
    using SharedConstPtr = std::shared_ptr<const A_SVGF>;
    void load() override;

    void render(Scene::SharedPtr& pScene, Samples::SharedPtr& pSamples, Fbo::SharedPtr& pSourceFBO, Fbo::SharedPtr& pTargetFBO) override;

    void shutdown() override;

    void initUI() override;
    void destroyUI() override;
    void resizeSwapChain(const glm::ivec2& imgSize) override;

    void resetState() override;

    std::string getFilterName() override { return "A-SVGF"; }
    void A_SVGF::create_diff_fbo();

    bool       mEnable            = true;
    bool       mModulateAlbedo    = true;
    int        mNumIterations     = 5;
    int        mHistoryTap        = 0;
    int        mFilterKernel      = 1;  
    float      mTemporalAlpha = 0.1f;
    int        mDiffAtrousIterations = 5;
    int        mGradientFilterRadius = 2;
    bool       mNormalizeGradient = true;

    bool       mShowAntilagAlpha  = false;

    SERIALIZATION
    {
        SERIALIZE_VARIABLE(mEnable);
        SERIALIZE_VARIABLE(mModulateAlbedo);
        SERIALIZE_VARIABLE(mNumIterations);
        SERIALIZE_VARIABLE(mHistoryTap);
        SERIALIZE_VARIABLE(mTemporalAlpha);
        SERIALIZE_VARIABLE(mFilterKernel);
        SERIALIZE_VARIABLE(mDiffAtrousIterations);
        SERIALIZE_VARIABLE(mShowAntilagAlpha);
        SERIALIZE_VARIABLE(mGradientFilterRadius);
        SERIALIZE_VARIABLE(mNormalizeGradient);

        // ugly hack for evaluation, variable not part of struct!
        SERIALIZE_VARIABLE(gradientDownsample);
    }
private:
#define _SVGF_SHADER(a) \
    FullScreenPass::UniquePtr   mpPrg##a; \
    UniformBuffer::SharedPtr    mpPrg##a##CB;

    _SVGF_SHADER(TemporalAccumulation)
    _SVGF_SHADER(EstimateVariance)
    _SVGF_SHADER(Atrous)
    _SVGF_SHADER(CreateGradientSamples)
    _SVGF_SHADER(AtrousGradient)
#undef _SVGF_SHADER

    Fbo::SharedPtr              mpFBOAccum, mpFBOAccumPrev, mpFBOPing, mpFBOPong, mpColorHistory;
    Fbo::SharedPtr              mpColorHistoryUnfiltered, mpAntilagAlpha;
    Fbo::SharedPtr              mpDiffPing, mpDiffPong;
};
